/*
Project Gutenberg HTML docutils stylesheet.

This stylesheet contains styles specific to HTML.
*/

/* FONTS */

/* em                { font-style: normal }
strong            { font-weight: normal } */

.small-caps       { font-variant: small-caps }
.gesperrt         { letter-spacing: 0.1em }

/* ALIGN */

.align-left       { clear: left;
		    float: left;
		    margin-right: 1em }

.align-right      { clear: right;
		    float: right;
		    margin-left: 1em }

.align-center     { margin-left: auto;
		    margin-right: auto }

div.shrinkwrap    { display: table; }

/* SECTIONS */

body              { margin: 5% 10% 5% 10% }

/* compact list items containing just one p */
li p.pfirst       { margin-top: 0; margin-bottom: 0 }

.first            { margin-top: 0 !important;
		    text-indent: 0 !important }
.last             { margin-bottom: 0 !important }

span.dropcap      { float: left; margin: 0 0.1em 0 0; line-height: 1 }
img.dropcap       { float: left; margin: 0 0.5em 0 0; max-width: 25% }
span.dropspan     { font-variant: small-caps }

.no-page-break    { page-break-before: avoid !important }

/* PAGINATION */

.pageno           { position: absolute; right: 95%; font: medium sans-serif; text-indent: 0 }
.pageno:after     { color: gray; content: '[' attr(title) ']' }
.lineno           { position: absolute; left:  95%; font: medium sans-serif; text-indent: 0 }
.lineno:after     { color: gray; content: '[' attr(title) ']' }
.toc-pageref      { float: right }

@media screen {
   .coverpage, .frontispiece, .titlepage, .verso, .dedication, .plainpage
                       { margin: 10% 0; }

   div.clearpage, div.cleardoublepage
                       { margin: 10% 0; border: none; border-top: 1px solid gray; }

   .vfill              { margin:  5% 10% }
}

@media print {
   div.clearpage       { page-break-before: always; padding-top: 10% }
   div.cleardoublepage { page-break-before: right;  padding-top: 10%  }

   .vfill              { margin-top: 20% }
   h2.title            { margin-top: 20% }
}

/* DIV */
pre               { font-family: monospace; font-size: 0.9em; white-space: pre-wrap }
